from datetime import datetime
from typing import Optional

from pydantic import BaseModel, confloat, root_validator, validator


class DataPointBase(BaseModel, use_enum_values=True):
    """Data-point

    Contains single raw data value received on API.
    This is just base class - plain, observation or timeseries datapoints inherit from this class
    (see below).

    Provides front line of validation for this data value.

    Internal usage: inside Task, created by TaskExecutor
    """

    etype: str
    eid: str
    attr: str
    src: Optional[str] = None


class DataPointPlainBase(DataPointBase):
    """Plain attribute data-point

    Contains single raw data value received on API for plain attribute.

    In case of plain data-point, it's not really a data-point, but we use
    the same naming for simplicity.
    """


class DataPointObservationsBase(DataPointBase):
    """Observations attribute data-point

    Contains single raw data value received on API for observations attribute.
    """

    t1: datetime
    t2: Optional[datetime] = None
    c: confloat(ge=0.0, le=1.0) = 1.0

    @validator("t2", always=True)
    def validate_t2(cls, v, values):
        v = v or values.get("t1") or datetime.now()
        if "t1" in values:
            assert values["t1"] <= v, "'t2' is before 't1'"
        return v


class DataPointTimeseriesBase(DataPointBase):
    """Timeseries attribute data-point

    Contains single raw data value received on API for observations attribute.
    """

    t1: datetime
    t2: Optional[datetime] = None

    @validator("t2")
    def validate_t2(cls, v, values):
        if "t1" in values:
            assert values["t1"] <= v, "'t2' is before 't1'"
        return v


def is_list_ordered(to_check: list):
    """Checks if list is ordered (not decreasing anywhere)"""
    return all(to_check[i] <= to_check[i + 1] for i in range(len(to_check) - 1))


# Timeseries validators
# They are included by child classes of `DataPointTimeseriesBase` in `AttrSpecTimeseries`,
# where `v` is available
# I know, this is very dirty, but there's no other way around.
@validator("v")
def dp_ts_v_validator(cls, v, values):
    # Check if all value arrays are the same length
    values_len = [len(v_i) for _, v_i in v.dict().items()]
    assert len(set(values_len)) == 1, f"Series values have different lengths: {values_len}"

    return v


def dp_ts_root_validator_regular_wrapper(time_step):
    def dp_ts_root_validator_regular(cls, values):
        """Validates or sets t2 of regular timeseries datapoint"""
        if "v" in values and "t1" in values:
            # Get length of first value series included in datapoint
            first_series = list(values["v"].dict().keys())[0]
            series_len = len(values["v"].dict()[first_series])
            correct_t2 = values["t1"] + series_len * time_step

            if "t2" in values and values["t2"]:
                assert (
                    values["t2"] == correct_t2
                ), "Difference of t1 and t2 is invalid. Must be values_len*time_step."
            else:
                values["t2"] = correct_t2

        return values

    return root_validator(dp_ts_root_validator_regular, allow_reuse=True)


@root_validator
def dp_ts_root_validator_irregular(cls, values):
    """Validates or sets t2 of irregular timeseries datapoint"""
    if "v" in values:
        first_time = values["v"].time[0]
        last_time = values["v"].time[-1]

        # Check t1 <= first_time
        if "t1" in values:
            assert (
                values["t1"] <= first_time
            ), f"'t1' is above first item in 'time' series ({first_time})"

        # Check last_time <= t2
        if "t2" in values and values["t2"]:
            assert (
                values["t2"] >= last_time
            ), f"'t2' is below last item in 'time' series ({last_time})"
        else:
            values["t2"] = last_time

        # time must be ordered
        assert is_list_ordered(values["v"].time), "'time' series is not ordered"

    return values


@root_validator
def dp_ts_root_validator_irregular_intervals(cls, values):
    """Validates or sets t2 of irregular intervals timeseries datapoint"""
    if "v" in values:
        first_time = values["v"].time_first[0]
        last_time = values["v"].time_last[-1]

        # Check t1 <= first_time
        if "t1" in values:
            assert (
                values["t1"] <= first_time
            ), f"'t1' is above first item in 'time_first' series ({first_time})"

        # Check last_time <= t2
        if "t2" in values and values["t2"]:
            assert (
                values["t2"] >= last_time
            ), f"'t2' is below last item in 'time_last' series ({last_time})"
        else:
            values["t2"] = last_time

        # Check time_first[i] <= time_last[i]
        assert all(
            t[0] <= t[1] for t in zip(values["v"].time_first, values["v"].time_last)
        ), "'time_first[i] <= time_last[i]' isn't true for all 'i'"

    return values



from abc import ABC, abstractmethod

from dp3.common.callback_registrar import CallbackRegistrar
from dp3.common.config import PlatformConfig


class BaseModule(ABC):
    """
    Abstract class for platform modules.
    Every module must inherit this abstract class for automatic loading of module!
    """

    @abstractmethod
    def __init__(
        self, platform_config: PlatformConfig, module_config: dict, registrar: CallbackRegistrar
    ):
        """Initialize the module and register callbacks.
        Args:
            platform_config: Platform configuration class
            module_config: Configuration of the module,
                equivalent of `platform_config.config.get("modules.<module_name>")`
            registrar: A callback / hook registration interface
        """

    def start(self) -> None:
        """
        Run the module - used to run own thread if needed.

        Called after initialization, may be used to create and run a separate
        thread if needed by the module. Do nothing unless overridden.
        """
        return None

    def stop(self) -> None:
        """
        Stop the module - used to stop own thread.

        Called before program exit, may be used to finalize and stop the
        separate thread if it is used. Do nothing unless overridden.
        """
        return None



import logging

from pymongo import UpdateOne

from dp3.common.callback_registrar import CallbackRegistrar
from dp3.common.config import PlatformConfig
from dp3.common.datapoint import DataPointObservationsBase, DataPointTimeseriesBase
from dp3.common.task import DataPointTask
from dp3.database.database import EntityDatabase


class Telemetry:
    def __init__(
        self, db: EntityDatabase, platform_config: PlatformConfig, registrar: CallbackRegistrar
    ) -> None:
        self.log = logging.getLogger("Telemetry")

        self.db = db
        self.model_spec = platform_config.model_spec
        # self.config = platform_config.config.get("telemetry")  # No config for now

        # Schedule master document aggregation
        registrar.register_task_hook("on_task_start", self.note_latest_src_timestamp)

    def note_latest_src_timestamp(self, task: DataPointTask):
        cache_col = self.db.get_module_cache()
        updates = []
        for dp in task.data_points:
            has_timestamp = isinstance(dp, (DataPointObservationsBase, DataPointTimeseriesBase))
            if dp.src is None or not has_timestamp:
                self.log.debug("Skipping datapoint without src or timestamp: %s", dp)
                continue
            latest_timestamp = dp.t2 or dp.t1
            updates.append(
                UpdateOne(
                    {"_id": dp.src},
                    [{"$set": {"_id": dp.src, "src_t": {"$max": ["$src_t", latest_timestamp]}}}],
                    upsert=True,
                )
            )

        if not updates:
            return

        res = cache_col.bulk_write(updates)
        self.log.debug(
            "Updating %s src_timestamp records: %s modified",
            len(updates),
            res.modified_count,
        )



from fastapi import APIRouter

from dp3.api.internal.config import CONTROL_WRITER
from dp3.api.internal.response_models import SuccessResponse
from dp3.common.control import ControlAction, ControlMessage

router = APIRouter()


@router.get("/{action}")
async def execute_action(action: ControlAction) -> SuccessResponse:
    """Sends the given action into execution queue."""
    CONTROL_WRITER.put_task(ControlMessage(action=action))
    return SuccessResponse(detail="Action sent.")



from datetime import datetime
from typing import Optional

from fastapi import APIRouter, Depends, HTTPException, Request
from pydantic import NonNegativeInt, PositiveInt, ValidationError

from dp3.api.internal.config import DB, MODEL_SPEC, TASK_WRITER
from dp3.api.internal.entity_response_models import (
    EntityEidAttrValue,
    EntityEidAttrValueOrHistory,
    EntityEidData,
    EntityEidList,
)
from dp3.api.internal.helpers import api_to_dp3_datapoint
from dp3.api.internal.models import (
    DataPoint,
)
from dp3.api.internal.response_models import ErrorResponse, RequestValidationError, SuccessResponse
from dp3.common.attrspec import AttrType
from dp3.common.task import DataPointTask
from dp3.database.database import DatabaseError


async def check_etype(etype: str):
    """Middleware to check entity type existence"""
    if etype not in MODEL_SPEC.entities:
        raise RequestValidationError(["path", "etype"], f"Entity type '{etype}' doesn't exist")
    return etype


router = APIRouter(dependencies=[Depends(check_etype)])


@router.get(
    "/{etype}", responses={400: {"description": "Query can't be processed", "model": ErrorResponse}}
)
async def list_entity_type_eids(
    etype: str, eid_filter: str = "", skip: NonNegativeInt = 0, limit: PositiveInt = 20
) -> EntityEidList:
    """List latest snapshots of all `id`s present in database under `etype`.

    Contains only latest snapshot.

    Uses pagination.

    If `eid_filter` is not empty, returns only `id`s containing substring `eid_filter`.
    """
    try:
        cursor, total_count = DB.get_latest_snapshots(etype, eid_filter)
        cursor_page = cursor.skip(skip).limit(limit)
    except DatabaseError as e:
        raise HTTPException(status_code=400, detail="Query is invalid") from e

    time_created = None

    # Remove _id field
    result = list(cursor_page)
    for r in result:
        time_created = r["_time_created"]
        del r["_time_created"]
        del r["_id"]

    return EntityEidList(
        time_created=time_created, count=len(result), total_count=total_count, data=result
    )


@router.get("/{etype}/{eid}")
async def get_eid_data(
    etype: str, eid: str, date_from: Optional[datetime] = None, date_to: Optional[datetime] = None
) -> EntityEidData:
    """Get data of `etype`'s `eid`.

    Contains all snapshots and master record.
    Snapshots are ordered by ascending creation time.
    """
    # Get master record
    # TODO: This is probably not the most efficient way. Maybe gather only
    # plain data from master record and then call `get_timeseries_history`
    # for timeseries.
    master_record = DB.get_master_record(etype, eid)
    if "_id" in master_record:
        del master_record["_id"]
    if "#hash" in master_record:
        del master_record["#hash"]

    entity_attribs = MODEL_SPEC.attribs(etype)

    # Get filtered timeseries data
    for attr in master_record:
        # Check for no longer existing attributes
        if attr in entity_attribs and entity_attribs[attr].t == AttrType.TIMESERIES:
            master_record[attr] = DB.get_timeseries_history(
                etype, attr, eid, t1=date_from, t2=date_to
            )

    # Get snapshots
    snapshots = list(DB.get_snapshots(etype, eid, t1=date_from, t2=date_to))
    for s in snapshots:
        del s["_id"]

    # Whether this eid contains any data
    empty = not master_record and len(snapshots) == 0

    return EntityEidData(empty=empty, master_record=master_record, snapshots=snapshots)


@router.get("/{etype}/{eid}/get/{attr}")
async def get_eid_attr_value(
    etype: str,
    eid: str,
    attr: str,
    date_from: Optional[datetime] = None,
    date_to: Optional[datetime] = None,
) -> EntityEidAttrValueOrHistory:
    """Get attribute value

    Value is either of:
    - current value: in case of plain attribute
    - current value and history: in case of observation attribute
    - history: in case of timeseries attribute
    """
    # Check if attribute exists
    if attr not in MODEL_SPEC.attribs(etype):
        raise RequestValidationError(["path", "attr"], f"Attribute '{attr}' doesn't exist")

    value_or_history = DB.get_value_or_history(etype, attr, eid, t1=date_from, t2=date_to)

    return EntityEidAttrValueOrHistory(attr_type=MODEL_SPEC.attr(etype, attr).t, **value_or_history)


@router.post("/{etype}/{eid}/set/{attr}")
async def set_eid_attr_value(
    etype: str, eid: str, attr: str, body: EntityEidAttrValue, request: Request
) -> SuccessResponse:
    """Set current value of attribute

    Internally just creates datapoint for specified attribute and value.

    This endpoint is meant for `editable` plain attributes -- for direct user edit on DP3 web UI.
    """
    # Check if attribute exists
    if attr not in MODEL_SPEC.attribs(etype):
        raise RequestValidationError(["path", "attr"], f"Attribute '{attr}' doesn't exist")

    # Construct datapoint
    try:
        dp = DataPoint(
            type=etype,
            id=eid,
            attr=attr,
            v=body.value,
            t1=datetime.now(),
            src=f"{request.client.host} via API",
        )
        dp3_dp = api_to_dp3_datapoint(dp.dict())
    except ValidationError as e:
        raise RequestValidationError(["body", "value"], e.errors()[0]["msg"]) from e

    # This shouldn't fail
    task = DataPointTask(model_spec=MODEL_SPEC, etype=etype, eid=eid, data_points=[dp3_dp])

    # Push tasks to task queue
    TASK_WRITER.put_task(task, False)

    # Datapoints from this endpoint are intentionally not logged using `DPLogger`.
    # If for some reason, in the future, they need to be, just copy code from data ingestion
    # endpoint.

    return SuccessResponse()



"""
Core module for dp3 - adds NRU (next regular update) fields to the newly added entity.
Updater uses NRU fields to issue corresponding regular updates at the specified time.
Various modules may hook their functions to the regular updates.

Note:
    Core modules needs to be enabled in worker.py and not in config!
"""
import logging
from datetime import timedelta

from dp3 import g
from dp3.common.base_module import BaseModule


class UpdatePlanner(BaseModule):
    def __init__(self):
        # the name of logger should be the same as the name of module itself
        self.log = logging.getLogger("UpdatePlanner")
        # logging level depends on user (DEBUG is useful while the module is tested or implemented)
        self.log.setLevel("DEBUG")

        # register handler for every etype
        for etype in g.model_spec:
            g.td.register_handler(
                self.processing_function,  # function to call (callback)
                etype,  # entity type, which actions will be watched (e.g. "ip")
                # tuple/list/set of attributes to watch
                # (their update triggers call of the registered method)
                ("!NEW",),
                ("_nru1d", "_nru1w"),  # tuple/list/set of attributes the method may change
            )
            self.log.debug(f"Registred handler for {etype}.")

    def processing_function(self, etype, ekey, record, updates):
        """
        Args:
            etype: entity type (e.g. 'ip')
            ekey: entity identificator (e.g. "1.2.3.4")
            record: instance of Record as database record cache
            updates: list of all attributes whose update triggered this call and
                their new values (or events and their parameters) as a list of 3-tuples:
                [(attr, new_val, old_val), (!event, param), ...]
        Returns:
             new request updates
        """

        return [
            {"op": "set", "attr": "_nru1d", "val": record["ts_added"] + timedelta(days=1)},
            {"op": "set", "attr": "_nru1w", "val": record["ts_added"] + timedelta(days=7)},
        ]



from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from typing import Any, Optional

from pydantic import BaseModel, root_validator, validator
from pydantic.error_wrappers import ValidationError

from dp3.common.config import ModelSpec
from dp3.common.datapoint import DataPointBase


class Task(BaseModel, ABC):
    """
    A generic task type class.

    An abstraction for the [task queue][dp3.task_processing.task_queue] classes to depend upon.
    """

    @abstractmethod
    def routing_key(self) -> str:
        """
        Returns:
            A string to be used as a routing key between workers.
        """

    @abstractmethod
    def as_message(self) -> str:
        """
        Returns:
            A string representation of the object.
        """


class DataPointTask(Task):
    """DataPointTask

    Contains single task to be pushed to TaskQueue and processed.
    Attributes:
        etype: Entity type
        eid: Entity id / key
        data_points: List of DataPoints to process
        tags: List of tags
        ttl_token: ...
    """

    # Model specification just for internal validation. Discarded after that.
    model_spec: ModelSpec

    etype: str
    eid: str
    data_points: list[DataPointBase] = []
    tags: list[Any] = []
    ttl_token: Optional[datetime] = None

    def routing_key(self):
        return f"{self.etype}:{self.eid}"

    def as_message(self) -> str:
        return self.json(exclude={"model_spec"})

    @validator("etype")
    def validate_etype(cls, v, values):
        if "model_spec" in values:
            assert v in values["model_spec"], f"Invalid etype '{v}'"
        return v

    @validator("data_points", pre=True, each_item=True)
    def instanciate_dps(cls, v, values):
        if "model_spec" in values and values["model_spec"]:
            # Convert `DataPointBase` instances back to dicts
            if isinstance(v, DataPointBase):
                v = v.dict()

            etype = v.get("etype")
            attr = v.get("attr")

            # Fetch datapoint model
            try:
                dp_model = values["model_spec"].attr(etype, attr).dp_model
            except (KeyError, TypeError) as e:
                raise ValueError(f"Attribute '{attr}' not found in entity '{etype}'") from e

            # Parse datapoint using model from attribute specification
            try:
                return dp_model.parse_obj(v)
            except ValidationError as e:
                raise ValueError(e) from e

        return v

    @validator("data_points", each_item=True)
    def validate_data_points(cls, v, values):
        if "etype" in values and "eid" in values:
            assert (
                v.etype == values["etype"]
            ), f"Task's etype '{values['etype']}' doesn't match datapoint's etype '{v.etype}'"
            assert (
                v.eid == values["eid"]
            ), f"Task's eid '{values['eid']}' doesn't match datapoint's eid '{v.eid}'"
        return v

    @root_validator
    def discard_attr_spec(cls, values):
        # This is run at the end of validation.
        # Discard attribute specification.
        del values["model_spec"]
        return values


class SnapshotMessageType(Enum):
    task = "task"
    linked_entities = "linked_entities"


class Snapshot(Task):
    """Snapshot

    Contains a list of entities, the meaning of which depends on the `type`.
    If `type` is "task", then the list contains linked entities for which a snapshot
    should be created. Otherwise `type` is "linked_entities", indicating which entities
    must be skipped in a parallelized creation of unlinked entities.

    Attributes:
        entities: List of (entity_type, entity_id)
        time: timestamp for snapshot creation
    """

    entities: list[tuple[str, str]]
    time: datetime
    type: SnapshotMessageType

    def routing_key(self):
        return "-".join(f"{etype}:{eid}" for etype, eid in self.entities)

    def as_message(self) -> str:
        return self.json()



import ipaddress
import re
from datetime import datetime
from enum import Enum
from typing import Any, Union

from pydantic import BaseModel, Extra, Json, PrivateAttr, constr, create_model

# Regular expressions for parsing various data types
re_array = re.compile(r"^array<(\w+)>$")
re_set = re.compile(r"^set<(\w+)>$")
re_link = re.compile(r"^link\s*<\s*(?P<etype>\w+)\s*(?:,\s*(?P<data>\S+)\s*)?>$")
re_dict = re.compile(r"^dict\s*<(\s*(\w+\??:\w+,\s*)*(\w+\??:\w+\s*))>$")
re_category = re.compile(
    r"^category<\s*(?P<type>\w+)\s*;\s*(?P<vals>(?:\s*\w+,\s*)*(?:\s*\w+\s*))>$"
)

# Dictionary containing validator functions for primitive data types
primitive_data_types = {
    "tag": bool,
    "binary": bool,
    "string": str,
    "int": int,
    "int64": int,
    "float": float,
    "ipv4": ipaddress.IPv4Address,
    "ipv6": ipaddress.IPv6Address,
    "mac": constr(regex=r"^([0-9a-fA-F]{2}[:-]){5}([0-9a-fA-F]{2})$"),
    "time": datetime,
    "special": Any,
    "json": Union[Json[Any], dict, list],
}


class Link(BaseModel, extra=Extra.forbid):
    eid: str


class DataType(BaseModel):
    """Data type container

    Represents one of primitive data types:

    - tag
    - binary
    - string
    - int
    - int64
    - float
    - ipv4
    - ipv6
    - mac
    - time
    - special
    - json

    or composite data type:

    - link<entity_type>
    - array<str_type>
    - set<str_type>
    - dict<keys>
    - category<str_type;category1,category2,...>

    Attributes:
        data_type: type for incoming value validation
        hashable: whether contained data is hashable
        is_link: whether this data type is link
        link_to: if `is_link` is True, what is linked target
    """

    __root__: str
    _data_type = PrivateAttr(None)
    _hashable = PrivateAttr(True)
    _is_link = PrivateAttr(False)
    _link_to = PrivateAttr()
    _link_data = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)

        str_type = data["__root__"]

        self._hashable = not (
            "dict" in str_type
            or "set" in str_type
            or "array" in str_type
            or "special" in str_type
            or "json" in str_type
            or "link" in str_type
        )

        self.determine_value_validator(str_type)

    def determine_value_validator(self, str_type: str):
        """Determines value validator (inner `data_type`)

        This is not implemented inside `@validator`, because it apparently doesn't work with
        `__root__` models.
        """
        data_type = None

        if not isinstance(str_type, str):
            raise TypeError(f"Data type {str_type} is not string")

        if str_type in primitive_data_types:
            # Primitive type
            data_type = primitive_data_types[str_type]

        elif re.match(re_array, str_type):
            # Array
            element_type = str_type.split("<")[1].split(">")[0]
            if element_type not in primitive_data_types:
                raise TypeError(f"Data type {element_type} is not supported as an array element")
            data_type = list[primitive_data_types[element_type]]

        elif re.match(re_set, str_type):
            # Set
            element_type = str_type.split("<")[1].split(">")[0]
            if element_type not in primitive_data_types:
                raise TypeError(f"Data type {element_type} is not supported as an set element")
            data_type = list[primitive_data_types[element_type]]  # set is not supported by MongoDB

        elif m := re.match(re_link, str_type):
            # Link
            etype, data = m.group("etype"), m.group("data")
            self._link_to = etype
            self._is_link = True
            self._link_data = bool(data)

            if etype and data:
                value_type = DataType(__root__=data)
                data_type = create_model(
                    f"Link<{data}>", __base__=Link, data=(value_type._data_type, ...)
                )
            else:
                data_type = Link

        elif re.match(re_dict, str_type):
            # Dict
            dict_spec = {}

            key_str = str_type.split("<")[1].split(">")[0]
            key_spec = dict(item.strip().split(":") for item in key_str.split(","))

            # For each dict key
            for k, v in key_spec.items():
                if v not in primitive_data_types:
                    raise TypeError(f"Data type {v} of key {k} is not supported as a dict field")

                # Optional subattribute
                k_optional = k[-1] == "?"

                if k_optional:
                    # Remove question mark from key
                    k = k[:-1]

                # Set (type, default value) for the key
                dict_spec[k] = (primitive_data_types[v], None if k_optional else ...)

            # Create model for this dict
            data_type = create_model(f"{str_type}__inner", **dict_spec)

        elif m := re.match(re_category, str_type):
            # Category
            category_type, category_values = m.group("type"), m.group("vals")

            category_type = DataType(__root__=category_type)
            category_values = [
                category_type._data_type(value.strip()) for value in category_values.split(",")
            ]

            data_type = Enum(f"Category<{category_type}>", {val: val for val in category_values})
        else:
            raise TypeError(f"Data type '{str_type}' is not supported")

        # Set data type
        self._data_type = data_type

    @property
    def data_type(self) -> str:
        return self._data_type

    @property
    def hashable(self) -> bool:
        return self._hashable

    @property
    def is_link(self) -> bool:
        return self._is_link

    @property
    def link_to(self) -> str:
        return self._link_to

    def get_linked_entity(self) -> id:
        """Returns linked entity id. Raises ValueError if DataType is not a link."""
        try:
            return self._link_to
        except AttributeError:
            raise ValueError(f"DataType '{self}' is not a link.") from None

    def link_has_data(self) -> bool:
        """Whether link has data. Raises ValueError if DataType is not a link."""
        try:
            return self._link_data
        except AttributeError:
            raise ValueError(f"DataType '{self}' is not a link.") from None

    def __str__(self):
        return self.__root__

    def __repr__(self):
        return f"'{str(self)}'"



from pydantic import BaseModel, Extra


class SpecModel(BaseModel, extra=Extra.forbid):
    ...


class EntitySpec(SpecModel):
    """Entity specification

    This class represents specification of an entity type (e.g. ip, asn, ...)
    """

    id: str
    name: str
    snapshot: bool

    description: str = ""



"""
Module managing creation of snapshots, enabling data correlation and saving snapshots to DB.

- Snapshots are created periodically (user configurable period)

- When a snapshot is created, several things need to happen:
    - all registered timeseries processing modules must be called
      - this should result in `observations` or `plain` datapoints, which will be saved to db
        and forwarded in processing
    - current value must be computed for all observations
      - load relevant section of observation's history and perform configured history analysis.
        Result = plain values
    - load plain attributes saved in master collection
    - A record of described plain data makes a `profile`
    - Profile is additionally extended by related entities
    - Callbacks for data correlation and fusion should happen here
    - Save the complete results into database as snapshots
"""
import logging
from collections import defaultdict
from datetime import datetime, timedelta
from typing import Any, Callable

import pymongo.errors
from pydantic import BaseModel
from pymongo import ReplaceOne

from dp3.common.attrspec import (
    AttrSpecObservations,
    AttrType,
    ObservationsHistoryParams,
)
from dp3.common.config import CronExpression, PlatformConfig
from dp3.common.datapoint import DataPointBase
from dp3.common.scheduler import Scheduler
from dp3.common.task import DataPointTask, Snapshot, SnapshotMessageType
from dp3.database.database import EntityDatabase
from dp3.snapshots.snapshot_hooks import (
    SnapshotCorrelationHookContainer,
    SnapshotTimeseriesHookContainer,
)
from dp3.task_processing.task_executor import TaskExecutor
from dp3.task_processing.task_queue import TaskQueueReader, TaskQueueWriter

DB_SEND_CHUNK = 100
RETRY_COUNT = 3


class SnapShooterConfig(BaseModel):
    creation_rate: CronExpression = CronExpression(minute="*/30")


class SnapShooter:
    """Class responsible for creating entity snapshots."""

    def __init__(
        self,
        db: EntityDatabase,
        task_queue_writer: TaskQueueWriter,
        task_executor: TaskExecutor,
        platform_config: PlatformConfig,
        scheduler: Scheduler,
    ) -> None:
        self.log = logging.getLogger("SnapShooter")

        self.db = db
        self.task_queue_writer = task_queue_writer
        self.model_spec = platform_config.model_spec
        self.entity_relation_attrs = defaultdict(dict)
        for (entity, attr), _ in self.model_spec.relations.items():
            self.entity_relation_attrs[entity][attr] = True
        for entity in self.model_spec.entities:
            self.entity_relation_attrs[entity]["_id"] = True

        self.worker_index = platform_config.process_index
        self.worker_cnt = platform_config.num_processes
        self.config = SnapShooterConfig.parse_obj(platform_config.config.get("snapshots"))

        elog = task_executor.elog

        self._timeseries_hooks = SnapshotTimeseriesHookContainer(self.log, self.model_spec, elog)
        self._correlation_hooks = SnapshotCorrelationHookContainer(self.log, self.model_spec, elog)

        queue = f"{platform_config.app_name}-worker-{platform_config.process_index}-snapshots"
        self.snapshot_queue_reader = TaskQueueReader(
            callback=self.process_snapshot_task,
            parse_task=Snapshot.parse_raw,
            app_name=platform_config.app_name,
            worker_index=platform_config.process_index,
            rabbit_config=platform_config.config.get("processing_core.msg_broker", {}),
            queue=queue,
            priority_queue=queue,
            parent_logger=self.log,
        )

        self.snapshot_entities = [
            entity for entity, spec in self.model_spec.entities.items() if spec.snapshot
        ]
        self.log.info("Snapshots will be created for entities: %s", self.snapshot_entities)

        # Register snapshot cache
        for (entity, attr), spec in self.model_spec.relations.items():
            if spec.t == AttrType.PLAIN:
                task_executor.register_attr_hook(
                    "on_new_plain", self.add_to_link_cache, entity, attr
                )
            elif spec.t == AttrType.OBSERVATIONS:
                task_executor.register_attr_hook(
                    "on_new_observation", self.add_to_link_cache, entity, attr
                )

        if platform_config.process_index != 0:
            self.log.debug(
                "Snapshot task creation will be disabled in this worker to avoid race conditions."
            )
            self.snapshot_queue_writer = None
            return

        self.snapshot_queue_writer = TaskQueueWriter(
            platform_config.app_name,
            platform_config.num_processes,
            platform_config.config.get("processing_core.msg_broker"),
            f"{platform_config.app_name}-main-snapshot-exchange",
            parent_logger=self.log,
        )

        # Schedule snapshot period
        snapshot_cron = self.config.creation_rate.dict(exclude_none=True)
        scheduler.register(self.make_snapshots, **snapshot_cron)

    def start(self):
        """Connect to RabbitMQ and start consuming from TaskQueue."""
        self.log.info("Connecting to RabbitMQ")
        self.snapshot_queue_reader.connect()
        self.snapshot_queue_reader.check()  # check presence of needed queues
        if self.snapshot_queue_writer is not None:
            self.snapshot_queue_writer.connect()
            self.snapshot_queue_writer.check()  # check presence of needed exchanges

        self.snapshot_queue_reader.start()

    def stop(self):
        """Stop consuming from TaskQueue, disconnect from RabbitMQ."""
        self.snapshot_queue_reader.stop()

        if self.snapshot_queue_writer is not None:
            self.snapshot_queue_writer.disconnect()
        self.snapshot_queue_reader.disconnect()

    def register_timeseries_hook(
        self,
        hook: Callable[[str, str, list[dict]], list[DataPointTask]],
        entity_type: str,
        attr_type: str,
    ):
        """
        Registers passed timeseries hook to be called during snapshot creation.

        Binds hook to specified `entity_type` and `attr_type` (though same hook can be bound
        multiple times).

        Args:
            hook: `hook` callable should expect entity_type, attr_type and attribute
                history as arguments and return a list of `DataPointTask` objects.
            entity_type: specifies entity type
            attr_type: specifies attribute type

        Raises:
            ValueError: If entity_type and attr_type do not specify a valid timeseries attribute,
                a ValueError is raised.
        """
        self._timeseries_hooks.register(hook, entity_type, attr_type)

    def register_correlation_hook(
        self,
        hook: Callable[[str, dict], None],
        entity_type: str,
        depends_on: list[list[str]],
        may_change: list[list[str]],
    ):
        """
        Registers passed hook to be called during snapshot creation.

        Binds hook to specified entity_type (though same hook can be bound multiple times).

        `entity_type` and attribute specifications are validated, `ValueError` is raised on failure.

        Args:
            hook: `hook` callable should expect entity type as str
                and its current values, including linked entities, as dict
            entity_type: specifies entity type
            depends_on: each item should specify an attribute that is depended on
                in the form of a path from the specified entity_type to individual attributes
                (even on linked entities).
            may_change: each item should specify an attribute that `hook` may change.
                specification format is identical to `depends_on`.

        Raises:
            ValueError: On failure of specification validation.
        """
        self._correlation_hooks.register(hook, entity_type, depends_on, may_change)

    def add_to_link_cache(self, eid: str, dp: DataPointBase):
        """Adds the given entity,eid pair to the cache of all linked entitites."""
        cache = self.db.get_module_cache()
        etype_to = self.model_spec.relations[dp.etype, dp.attr].relation_to
        to_insert = [
            {
                "_id": f"{dp.etype}#{eid}",
                "etype": dp.etype,
                "eid": eid,
                "expire_at": datetime.now() + timedelta(days=2),
            },
            {
                "_id": f"{etype_to}#{dp.v.eid}",
                "etype": etype_to,
                "eid": dp.v.eid,
                "expire_at": datetime.now() + timedelta(days=2),
            },
        ]
        res = cache.bulk_write([ReplaceOne({"_id": x["_id"]}, x, upsert=True) for x in to_insert])
        self.log.debug("Cached %s linked entities: %s", len(to_insert), res.bulk_api_result)

    def make_snapshots(self):
        """Creates snapshots for all entities currently active in database."""
        time = datetime.now()

        # distribute list of possibly linked entities to all workers
        cached = self.get_cached_link_entity_ids()
        self.log.debug("Broadcasting %s cached linked entities", len(cached))
        self.snapshot_queue_writer.broadcast_task(
            task=Snapshot(entities=cached, time=time, type=SnapshotMessageType.linked_entities)
        )

        # Load links only for a reduced set of entities
        self.log.debug("Loading linked entities.")
        self.db.save_metadata(time, {"task_creation_start": time, "entities": 0, "components": 0})
        times = {}
        counts = {"entities": 0, "components": 0}
        try:
            linked_entities = self.get_linked_entities(time, cached)
            times["components_loaded"] = datetime.now()

            for linked_entities_component in linked_entities:
                counts["entities"] += len(linked_entities_component)
                counts["components"] += 1

                self.snapshot_queue_writer.put_task(
                    task=Snapshot(
                        entities=linked_entities_component, time=time, type=SnapshotMessageType.task
                    )
                )
        except pymongo.errors.CursorNotFound as err:
            self.log.exception(err)
        finally:
            times["task_creation_end"] = datetime.now()
            self.db.update_metadata(
                time,
                metadata=times,
                increase=counts,
            )

    def get_cached_link_entity_ids(self):
        cache = self.db.get_module_cache()
        result = cache.aggregate([{"$group": {"_id": "$etype", "eid": {"$addToSet": "$eid"}}}])
        return [(res_obj["_id"], eid) for res_obj in result for eid in res_obj["eid"]]

    def get_linked_entities(self, time: datetime, cached_linked_entities: list[tuple[str, str]]):
        """Get weakly connected components from entity graph."""
        visited_entities = set()
        entity_to_component = {}
        linked_components = []
        for etype, eid in cached_linked_entities:
            master_record = self.db.get_master_record(
                etype, eid, projection=self.entity_relation_attrs[etype]
            ) or {"_id": eid}
            if (etype, master_record["_id"]) not in visited_entities:
                # Get entities linked by current entity
                current_values = self.get_values_at_time(etype, master_record, time)
                linked_entities = self.load_linked_entity_ids(etype, current_values, time)

                # Set linked as visited
                visited_entities.update(linked_entities)

                # Update component
                have_component = linked_entities & set(entity_to_component.keys())
                if have_component:
                    for entity in have_component:
                        component = entity_to_component[entity]
                        component.update(linked_entities)
                        entity_to_component.update(
                            {entity: component for entity in linked_entities}
                        )
                        break
                else:
                    entity_to_component.update(
                        {entity: linked_entities for entity in linked_entities}
                    )
                    linked_components.append(linked_entities)
        return linked_components

    def process_snapshot_task(self, msg_id, task: Snapshot):
        """
        Acknowledges the received message and makes a snapshot according to the `task`.

        This function should not be called directly, but set as callback for TaskQueueReader.
        """
        self.snapshot_queue_reader.ack(msg_id)
        if task.type == SnapshotMessageType.task:
            self.make_snapshot(task)
        elif task.type == SnapshotMessageType.linked_entities:
            self.make_snapshots_by_hash(task)
        else:
            raise ValueError("Unknown SnapshotMessageType.")

    def make_snapshots_by_hash(self, task: Snapshot):
        """
        Make snapshots for all entities with routing key belonging to this worker.
        """
        self.log.debug("Creating snapshots for worker portion by hash.")
        have_links = set(task.entities)
        entity_cnt = 0
        for etype in self.snapshot_entities:
            records_cursor = self.db.get_worker_master_records(
                self.worker_index,
                self.worker_cnt,
                etype,
                no_cursor_timeout=True,
            )
            for attempt in range(RETRY_COUNT):
                try:
                    entity_cnt += self.make_linkless_snapshots(
                        etype, records_cursor, task.time, have_links
                    )
                except Exception as err:
                    self.log.exception("Uncaught exception while creating snapshots: %s", err)
                    if attempt < RETRY_COUNT - 1:
                        self.log.info("Retrying snapshot creation for '%s' due to errors.", etype)
                    continue
                finally:
                    records_cursor.close()
                break
            else:
                self.log.error(
                    "Failed to create snapshots for '%s' after %s attempts.", etype, attempt + 1
                )
        self.db.update_metadata(
            task.time,
            metadata={},
            increase={"entities": entity_cnt, "components": entity_cnt},
        )
        self.log.debug("Worker snapshot creation done.")

    def make_linkless_snapshots(
        self, etype: str, records_cursor, time: datetime, have_links: set
    ) -> int:
        entity_cnt = 0
        snapshots = []
        for master_record in records_cursor:
            if (etype, master_record["_id"]) in have_links:
                continue
            entity_cnt += 1
            snapshots.append(self.make_linkless_snapshot(etype, master_record, time))
            if len(snapshots) >= DB_SEND_CHUNK:
                self.db.save_snapshots(etype, snapshots, time)
                snapshots.clear()

        if snapshots:
            self.db.save_snapshots(etype, snapshots, time)
            snapshots.clear()
        return entity_cnt

    def make_linkless_snapshot(self, entity_type: str, master_record: dict, time: datetime):
        """
        Make a snapshot for given entity `master_record` and `time`.

        Runs timeseries and correlation hooks.
        The resulting snapshot is saved into DB.
        """
        self.run_timeseries_processing(entity_type, master_record)
        values = self.get_values_at_time(entity_type, master_record, time)
        entity_values = {(entity_type, master_record["_id"]): values}

        self._correlation_hooks.run(entity_values)

        assert len(entity_values) == 1, "Expected a single entity."
        for record in entity_values.values():
            return record

    def make_snapshot(self, task: Snapshot):
        """
        Make a snapshot for entities and time specified by `task`.

        Runs timeseries and correlation hooks.
        The resulting snapshots are saved into DB.
        """
        entity_values = {}
        for entity_type, entity_id in task.entities:
            record = self.db.get_master_record(entity_type, entity_id) or {"_id": entity_id}
            self.run_timeseries_processing(entity_type, record)
            values = self.get_values_at_time(entity_type, record, task.time)
            entity_values[entity_type, entity_id] = values

        self.link_loaded_entities(entity_values)
        self._correlation_hooks.run(entity_values)

        # unlink entities again
        for rtype_rid, record in entity_values.items():
            rtype, rid = rtype_rid
            for attr, value in record.items():
                if (rtype, attr) not in self.model_spec.relations:
                    continue
                if (
                    self.model_spec.relations[rtype, attr].t == AttrType.OBSERVATIONS
                    and self.model_spec.relations[rtype, attr].multi_value
                ):
                    record[attr] = [
                        {k: v for k, v in link_dict.items() if k != "record"} for link_dict in value
                    ]
                else:
                    record[attr] = {k: v for k, v in value.items() if k != "record"}

        for rtype_rid, record in entity_values.items():
            self.db.save_snapshot(rtype_rid[0], record, task.time)

    def run_timeseries_processing(self, entity_type, master_record):
        """
        - all registered timeseries processing modules must be called
          - this should result in `observations` or `plain` datapoints, which will be saved to db
            and forwarded in processing
        """
        tasks = []
        for attr, attr_spec in self.model_spec.entity_attributes[entity_type].items():
            if attr_spec.t == AttrType.TIMESERIES and attr in master_record:
                new_tasks = self._timeseries_hooks.run(entity_type, attr, master_record[attr])
                tasks.extend(new_tasks)

        self.extend_master_record(entity_type, master_record, tasks)
        for task in tasks:
            self.task_queue_writer.put_task(task)

    @staticmethod
    def extend_master_record(etype, master_record, new_tasks: list[DataPointTask]):
        """Update existing master record with datapoints from new tasks"""
        for task in new_tasks:
            for datapoint in task.data_points:
                if datapoint.etype != etype:
                    continue
                dp_dict = datapoint.dict(include={"v", "t1", "t2", "c"})
                if datapoint.attr in master_record:
                    master_record[datapoint.attr].append()
                else:
                    master_record[datapoint.attr] = [dp_dict]

    def get_values_at_time(self, etype: str, master_record: dict, time: datetime) -> dict:
        current_values = {"eid": master_record["_id"]}
        for attr, attr_spec in self.model_spec.entity_attributes[etype].items():
            if (
                attr not in master_record
                or attr_spec.t not in AttrType.PLAIN | AttrType.OBSERVATIONS
            ):
                continue
            if attr_spec.t == AttrType.PLAIN:
                current_values[attr] = master_record[attr]["v"]
                continue

            if attr_spec.multi_value:
                val, conf = self.get_multi_value_at_time(attr_spec, master_record[attr], time)
            else:
                val, conf = self.get_value_at_time(attr_spec, master_record[attr], time)

            if conf:  # conf != 0.0 or len(conf) > 0
                current_values[attr] = val
                current_values[f"{attr}#c"] = conf
        return current_values

    def load_linked_entity_ids(self, entity_type: str, current_values: dict, time: datetime):
        """
        Loads the subgraph of entities linked to the current entity,
        returns a list of their types and ids.
        """
        loaded_entity_ids = {(entity_type, current_values["eid"])}
        linked_entity_ids_to_process = (
            self.get_linked_entity_ids(entity_type, current_values) - loaded_entity_ids
        )

        while linked_entity_ids_to_process:
            entity_identifiers = linked_entity_ids_to_process.pop()
            linked_etype, linked_eid = entity_identifiers
            relevant_attributes = self.entity_relation_attrs[linked_etype]
            record = self.db.get_master_record(
                linked_etype, linked_eid, projection=relevant_attributes
            ) or {"_id": linked_eid}
            linked_values = self.get_values_at_time(linked_etype, record, time)

            linked_entity_ids_to_process.update(
                self.get_linked_entity_ids(entity_type, linked_values) - set(loaded_entity_ids)
            )
            loaded_entity_ids.add((linked_etype, linked_eid))

        return loaded_entity_ids

    def get_linked_entity_ids(self, entity_type: str, current_values: dict) -> set[tuple[str, str]]:
        """
        Returns a set of tuples (entity_type, entity_id) identifying entities linked by
        `current_values`.
        """
        related_entity_ids = set()
        for attr, val in current_values.items():
            if (entity_type, attr) not in self.model_spec.relations:
                continue
            attr_spec = self.model_spec.relations[entity_type, attr]
            if attr_spec.t == AttrType.OBSERVATIONS and attr_spec.multi_value:
                related_entity_ids.update((attr_spec.relation_to, v["eid"]) for v in val)
            else:
                related_entity_ids.add((attr_spec.relation_to, val["eid"]))
        return related_entity_ids

    def link_loaded_entities(self, loaded_entities: dict):
        for identifiers, entity in loaded_entities.items():
            entity_type, entity_id = identifiers
            for attr, val in entity.items():
                if (entity_type, attr) not in self.model_spec.relations:
                    continue
                attr_spec = self.model_spec.relations[entity_type, attr]
                if attr_spec.t == AttrType.OBSERVATIONS and attr_spec.multi_value:
                    entity[attr] = [
                        {
                            **v,
                            "record": loaded_entities.get(
                                (attr_spec.relation_to, v["eid"]), {"eid": v["eid"]}
                            ),
                        }
                        for v in val
                    ]
                else:
                    entity[attr] = {
                        **val,
                        "record": loaded_entities.get(
                            (attr_spec.relation_to, val["eid"]), {"eid": val["eid"]}
                        ),
                    }

    def get_value_at_time(
        self, attr_spec: AttrSpecObservations, attr_history, time: datetime
    ) -> tuple[Any, float]:
        """Get current value of an attribute from its history. Assumes `multi_value = False`."""
        return max(
            (
                (point["v"], self.extrapolate_confidence(point, time, attr_spec.history_params))
                for point in attr_history
            ),
            key=lambda val_conf: val_conf[1],
            default=(None, 0.0),
        )

    def get_multi_value_at_time(
        self, attr_spec: AttrSpecObservations, attr_history, time: datetime
    ) -> tuple[list, list[float]]:
        """Get current value of a multi_value attribute from its history."""
        if attr_spec.data_type.hashable:
            values_with_confidence = defaultdict(float)
            for point in attr_history:
                value = point["v"]
                confidence = self.extrapolate_confidence(point, time, attr_spec.history_params)
                if confidence > 0.0 and values_with_confidence[value] < confidence:
                    values_with_confidence[value] = confidence
            return list(values_with_confidence.keys()), list(values_with_confidence.values())
        else:
            values = []
            confidence_list = []
            for point in attr_history:
                value = point["v"]
                confidence = self.extrapolate_confidence(point, time, attr_spec.history_params)
                if value in values:
                    i = values.index(value)
                    if confidence_list[i] < confidence:
                        confidence_list[i] = confidence
                elif confidence > 0.0:
                    values.append(value)
                    confidence_list.append(confidence)
            return values, confidence_list

    @staticmethod
    def extrapolate_confidence(
        datapoint: dict, time: datetime, history_params: ObservationsHistoryParams
    ) -> float:
        """Get the confidence value at given time."""
        t1 = datapoint["t1"]
        t2 = datapoint["t2"]
        base_confidence = datapoint["c"]

        if time < t1:
            if time <= t1 - history_params.pre_validity:
                return 0.0
            return base_confidence * (1 - (t1 - time) / history_params.pre_validity)
        if time <= t2:
            return base_confidence  # completely inside the (strict) interval
        if time >= t2 + history_params.post_validity:
            return 0.0
        return base_confidence * (1 - (time - t2) / history_params.post_validity)



"""Simple datapoint sender script for testing local DP3 instance."""

import json
import logging
import os
from argparse import ArgumentParser
from datetime import datetime
from itertools import islice

import pandas as pd
import requests


def get_valid_path(parser, arg):
    if not os.path.exists(arg):
        parser.error("The file %s does not exist!" % arg)
    else:
        return os.path.abspath(arg)


def get_datapoint_from_row(row):
    dp = dict(row.items())
    dp["t1"] = dp["t1"].strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]
    dp["t2"] = dp["t2"].strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]
    return dp


def get_shifted_datapoint_from_row(row):
    dp = dict(row.items())
    duration = dp["t2"] - dp["t1"]
    dp["t1"] = (datetime.utcnow()).strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]
    dp["t2"] = (datetime.utcnow() + duration).strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]
    return dp


def batched(iterable, n):
    "Batch data into tuples of length n. The last batch may be shorter."
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := list(islice(it, n)):
        yield batch


def cherry_pick_send(datapoints, dp_factory, log, args):
    # Get all datapoints that will be sent.
    attributes = datapoints["attr"].unique()
    log.debug("Found %s attributes", len(datapoints))
    log.debug(attributes)
    dp_list = []
    for attr in attributes:
        selected = datapoints[datapoints["attr"] == attr]
        selected_dps = selected.iloc[: args.count, :].apply(dp_factory, axis=1).to_list()
        log.debug("Selected %s datapoints of attribute %s", len(selected_dps), attr)
        dp_list.extend(selected_dps)

    log.info("Selected %s datapoints", len(dp_list))
    return dp_list


def send_all(datapoints, dp_factory, *_):
    for _, dp in datapoints.iterrows():
        yield dp_factory(dp)


if __name__ == "__main__":
    parser = ArgumentParser("Simple datapoint sender script for testing local DP3 instance.")
    parser.add_argument(
        "input_file",
        help="DataPoint log file in JSON format (pandas orient=records)",
        type=lambda x: get_valid_path(parser, x),
    )
    parser.add_argument(
        "mode",
        help="Mode of operation for the sender. "
        "Either cherry-pick datapoints by attribute, or send the entire log.",
        choices=["cherry-pick", "all"],
    )
    parser.add_argument(
        "--endpoint-url",
        help="base DP3 endpoint, default='http://127.0.0.1:5000'",
        default="http://127.0.0.1:5000",
        type=str,
        dest="endpoint_url",
    )
    parser.add_argument(
        "-n",
        dest="count",
        help="number of DataPoints to send per attribute, default=50",
        default=50,
        type=int,
    )
    parser.add_argument(
        "-c",
        "--chunk",
        dest="chunk",
        help="number of DataPoints to send per request, default=same as count",
        default=None,
        type=int,
    )
    parser.add_argument(
        "--shift-time",
        help="shift timestamps so that 't1' is the current UTC time",
        default=False,
        action="store_true",
        dest="shift_time",
    )
    args = parser.parse_args()
    if args.chunk is None:
        args.chunk = args.count
    dp_factory = get_shifted_datapoint_from_row if args.shift_time else get_datapoint_from_row
    logging.basicConfig(level=logging.INFO, format="%(name)s [%(levelname)s] %(message)s")
    log = logging.getLogger("DataPointSender")

    datapoints = pd.read_json(
        args.input_file,
        orient="records",
        convert_dates=["t1", "t2"],
    ).fillna(value={"c": 1.0})
    log.info("Input file contains %s DataPoints", datapoints.shape[0])

    if args.mode == "cherry-pick":
        dps = cherry_pick_send(datapoints, dp_factory, log, args)
    elif args.mode == "all":
        dps = send_all(datapoints, dp_factory, log, args)
    else:
        raise ValueError(f"Invalid mode selected: {args.mode}")

    # Send them
    datapoints_url = f"{args.endpoint_url}/datapoints"
    for batch in batched(dps, args.chunk):
        payload = json.dumps(batch)
        log.debug(payload)
        try:
            response = requests.post(url=datapoints_url, json=batch)
            if response.status_code == requests.codes.ok:
                attributes = {dp["attr"] for dp in batch}
                log.info("%s datapoints of attribute(s) %s OK", len(batch), ", ".join(attributes))
            else:
                log.error("Payload: %s", payload)
                log.error(
                    "Request failed: %s: %s", response.reason, response.content.decode("utf-8")
                )
        except requests.exceptions.ConnectionError as err:
            log.exception(err)
        except requests.exceptions.InvalidJSONError as err:
            log.exception(err)
            log.error(batch)



import json

from fastapi import FastAPI
from fastapi.exception_handlers import request_validation_exception_handler
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.cors import CORSMiddleware

from dp3.api.internal.config import CONFIG, DATAPOINTS_INGESTION_URL_PATH, DP_LOGGER, ROOT_PATH
from dp3.api.routers import control, entity, root

# Create new FastAPI app
app = FastAPI(root_path=ROOT_PATH)


# Redefine Pydantic validation error handler to log all bad datapoints
@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request, exc):
    if request.url.path.replace(ROOT_PATH, "") == DATAPOINTS_INGESTION_URL_PATH:
        # Convert body to readable form
        body = json.dumps(exc.body) if isinstance(exc.body, (dict, list)) else str(exc.body)

        DP_LOGGER.log_bad(body, str(exc), src=request.client.host)

    # Call default handler
    return await request_validation_exception_handler(request, exc)


# CORS
app.add_middleware(
    CORSMiddleware,
    allow_origins=CONFIG.get("api.cors.allow_origins", ["*"]),
    allow_methods=["*"],
    allow_headers=["*"],
)

# Register routers
app.include_router(entity.router, prefix="/entity", tags=["Entity"])
app.include_router(control.router, prefix="/control", tags=["Control"])
app.include_router(root.router)



from typing import Callable, Union

from dp3.common.scheduler import Scheduler
from dp3.common.task import DataPointTask
from dp3.snapshots.snapshooter import SnapShooter
from dp3.task_processing.task_executor import TaskExecutor


class CallbackRegistrar:
    """Interface for callback registration."""

    def __init__(
        self, scheduler: Scheduler, task_executor: TaskExecutor, snap_shooter: SnapShooter
    ):
        self._scheduler = scheduler
        self._task_executor = task_executor
        self._snap_shooter = snap_shooter

    def scheduler_register(
        self,
        func: Callable,
        *,
        func_args: Union[list, tuple] = None,
        func_kwargs: dict = None,
        year: Union[int, str] = None,
        month: Union[int, str] = None,
        day: Union[int, str] = None,
        week: Union[int, str] = None,
        day_of_week: Union[int, str] = None,
        hour: Union[int, str] = None,
        minute: Union[int, str] = None,
        second: Union[int, str] = None,
        timezone: str = "UTC",
    ) -> int:
        """
        Register a function to be run at specified times.

        Pass cron-like specification of when the function should be called,
        see [docs](https://apscheduler.readthedocs.io/en/latest/modules/triggers/cron.html)
        of apscheduler.triggers.cron for details.
        `
        Args:
            func: function or method to be called
            func_args: list of positional arguments to call func with
            func_kwargs: dict of keyword arguments to call func with
            year: 4-digit year
            month: month (1-12)
            day: day of month (1-31)
            week: ISO week (1-53)
            day_of_week: number or name of weekday (0-6 or mon,tue,wed,thu,fri,sat,sun)
            hour: hour (0-23)
            minute: minute (0-59)
            second: second (0-59)
            timezone: Timezone for time specification (default is UTC).
        Returns:
             job ID
        """
        return self._scheduler.register(
            func,
            func_args=func_args,
            func_kwargs=func_kwargs,
            year=year,
            month=month,
            day=day,
            week=week,
            day_of_week=day_of_week,
            hour=hour,
            minute=minute,
            second=second,
            timezone=timezone,
        )

    def register_task_hook(self, hook_type: str, hook: Callable):
        """Registers one of available task hooks

        See: [`TaskGenericHooksContainer`][dp3.task_processing.task_hooks.TaskGenericHooksContainer]
        in `task_hooks.py`
        """
        self._task_executor.register_task_hook(hook_type, hook)

    def register_entity_hook(self, hook_type: str, hook: Callable, entity: str):
        """Registers one of available task entity hooks

        See: [`TaskEntityHooksContainer`][dp3.task_processing.task_hooks.TaskEntityHooksContainer]
        in `task_hooks.py`
        """
        self._task_executor.register_entity_hook(hook_type, hook, entity)

    def register_attr_hook(self, hook_type: str, hook: Callable, entity: str, attr: str):
        """Registers one of available task attribute hooks

        See: [`TaskAttrHooksContainer`][dp3.task_processing.task_hooks.TaskAttrHooksContainer]
        in `task_hooks.py`
        """
        self._task_executor.register_attr_hook(hook_type, hook, entity, attr)

    def register_timeseries_hook(
        self,
        hook: Callable[[str, str, list[dict]], list[DataPointTask]],
        entity_type: str,
        attr_type: str,
    ):
        """
        Registers passed timeseries hook to be called during snapshot creation.

        Binds hook to specified `entity_type` and `attr_type` (though same hook can be bound
        multiple times).

        Args:
            hook: `hook` callable should expect entity_type, attr_type and attribute
                history as arguments and return a list of `DataPointTask` objects.
            entity_type: specifies entity type
            attr_type: specifies attribute type

        Raises:
            ValueError: If entity_type and attr_type do not specify a valid timeseries attribute,
                a ValueError is raised.
        """
        self._snap_shooter.register_timeseries_hook(hook, entity_type, attr_type)

    def register_correlation_hook(
        self,
        hook: Callable[[str, dict], None],
        entity_type: str,
        depends_on: list[list[str]],
        may_change: list[list[str]],
    ):
        """
        Registers passed hook to be called during snapshot creation.

        Binds hook to specified entity_type (though same hook can be bound multiple times).

        `entity_type` and attribute specifications are validated, `ValueError` is raised on failure.

        Args:
            hook: `hook` callable should expect entity type as str
                and its current values, including linked entities, as dict
            entity_type: specifies entity type
            depends_on: each item should specify an attribute that is depended on
                in the form of a path from the specified entity_type to individual attributes
                (even on linked entities).
            may_change: each item should specify an attribute that `hook` may change.
                specification format is identical to `depends_on`.

        Raises:
            ValueError: On failure of specification validation.
        """
        self._snap_shooter.register_correlation_hook(hook, entity_type, depends_on, may_change)



"""
Module enabling remote control of the platform's internal events.
"""
import logging
from enum import Enum
from typing import Callable

from pydantic import BaseModel

from dp3.common.config import PlatformConfig
from dp3.common.task import Task
from dp3.task_processing.task_queue import TaskQueueReader


class ControlAction(Enum):
    make_snapshots = "make_snapshots"


class ControlConfig(BaseModel):
    allowed_actions: list[ControlAction]


class ControlMessage(Task):
    action: ControlAction

    def routing_key(self):
        return ""

    def as_message(self) -> str:
        return self.json()


class Control:
    """Class enabling remote control of the platform's internal events."""

    def __init__(
        self,
        platform_config: PlatformConfig,
    ) -> None:
        self.log = logging.getLogger("Control")
        self.action_handlers: dict[ControlAction, Callable] = {}
        self.enabled = False

        if platform_config.process_index != 0:
            self.log.debug("Control will be disabled in this worker to avoid race conditions.")
            return

        self.enabled = True
        self.config = ControlConfig.parse_obj(platform_config.config.get("control"))
        self.allowed_actions = set(self.config.allowed_actions)
        self.log.debug("Allowed actions: %s", self.allowed_actions)

        queue = f"{platform_config.app_name}-control"
        self.control_queue = TaskQueueReader(
            callback=self.process_control_task,
            parse_task=ControlMessage.parse_raw,
            app_name=platform_config.app_name,
            worker_index=platform_config.process_index,
            rabbit_config=platform_config.config.get("processing_core.msg_broker", {}),
            queue=queue,
            priority_queue=queue,
            parent_logger=self.log,
        )

    def start(self):
        """Connect to RabbitMQ and start consuming from TaskQueue."""
        if not self.enabled:
            return

        unconfigured_handlers = self.allowed_actions - set(self.action_handlers)
        if unconfigured_handlers:
            raise ValueError(
                f"The following configured actions are missing handlers: {unconfigured_handlers}"
            )

        self.log.info("Connecting to RabbitMQ")
        self.control_queue.connect()
        self.control_queue.check()  # check presence of needed queues
        self.control_queue.start()

        self.log.debug("Configured handlers: %s", self.action_handlers)

    def stop(self):
        """Stop consuming from TaskQueue, disconnect from RabbitMQ."""
        if not self.enabled:
            return

        self.control_queue.stop()
        self.control_queue.disconnect()

    def set_action_handler(self, action: ControlAction, handler: Callable):
        """Sets the handler for the given action"""
        self.log.debug("Setting handler for action %s: %s", action, handler)
        self.action_handlers[action] = handler

    def process_control_task(self, msg_id, task: ControlMessage):
        """
        Acknowledges the received message and executes an action according to the `task`.

        This function should not be called directly, but set as callback for TaskQueueReader.
        """
        self.control_queue.ack(msg_id)
        if task.action in self.allowed_actions:
            self.log.info("Executing action: %s", task.action)
            self.action_handlers[task.action]()
        else:
            self.log.error("Action not allowed: %s", task.action)



import inspect
import logging
import time
import urllib
from datetime import datetime
from typing import Literal, Optional, Union

import pymongo
from pydantic import BaseModel, Field, validator
from pymongo import ReplaceOne
from pymongo.errors import OperationFailure

from dp3.common.attrspec import AttrType, timeseries_types
from dp3.common.config import HierarchicalDict, ModelSpec
from dp3.common.datapoint import DataPointBase
from dp3.task_processing.task_queue import HASH


class DatabaseError(Exception):
    pass


class MissingTableError(DatabaseError):
    pass


class MongoHostConfig(BaseModel, extra="forbid"):
    """MongoDB host."""

    address: str = "localhost"
    port: int = 27017


class MongoStandaloneConfig(BaseModel, extra="forbid"):
    """MongoDB standalone configuration."""

    mode: Literal["standalone"]
    host: MongoHostConfig = MongoHostConfig()


class MongoReplicaConfig(BaseModel, extra="forbid"):
    """MongoDB replica set configuration."""

    mode: Literal["replica"]
    replica_set: str = "dp3"
    hosts: list[MongoHostConfig]


class MongoConfig(BaseModel, extra="forbid"):
    """Database configuration."""

    db_name: str = "dp3"
    username: str = "dp3"
    password: str = "dp3"
    connection: Union[MongoStandaloneConfig, MongoReplicaConfig] = Field(..., discriminator="mode")

    @validator("username", "password")
    def url_safety(cls, v):
        return urllib.parse.quote_plus(v)


def get_caller_id():
    """Returns the name of the caller method's class, or function name if caller is not a method."""
    caller = inspect.stack()[2]
    if module := caller.frame.f_locals.get("self"):
        return module.__class__.__qualname__
    return caller.function


# number of seconds to wait for the i-th attempt to reconnect after error
RECONNECT_DELAYS = [1, 2, 5, 10, 30]


class EntityDatabase:
    """
    MongoDB database wrapper responsible for whole communication with database server.
    Initializes database schema based on database configuration.

    db_conf - configuration of database connection (content of database.yml)
    model_spec - ModelSpec object, configuration of data model (entities and attributes)
    """

    def __init__(
        self,
        db_conf: HierarchicalDict,
        model_spec: ModelSpec,
    ) -> None:
        self.log = logging.getLogger("EntityDatabase")

        config = MongoConfig.parse_obj(db_conf)

        self.log.info("Connecting to database...")
        for attempt, delay in enumerate(RECONNECT_DELAYS):
            try:
                self._db = self.connect(config)
                # Check if connected
                self._db.admin.command("ping")
            except pymongo.errors.ConnectionFailure as e:
                if attempt + 1 == len(RECONNECT_DELAYS):
                    raise DatabaseError(
                        "Cannot connect to database with specified connection arguments."
                    ) from e
                else:
                    self.log.error(
                        "Cannot connect to database (attempt %d, retrying in %ds).",
                        attempt + 1,
                        delay,
                    )
                    time.sleep(delay)

        self._db_schema_config = model_spec

        # Init and switch to correct database
        self._db = self._db[config.db_name]
        self._init_database_schema(config.db_name)

        self.log.info("Database successfully initialized!")

    @staticmethod
    def connect(config: MongoConfig) -> pymongo.MongoClient:
        if isinstance(config.connection, MongoStandaloneConfig):
            host = config.connection.host
            return pymongo.MongoClient(
                f"mongodb://{config.username}:{config.password}@{host.address}:{host.port}/",
                connectTimeoutMS=3000,
                serverSelectionTimeoutMS=5000,
            )
        elif isinstance(config.connection, MongoReplicaConfig):
            uri = (
                f"mongodb://{config.username}:{config.password}@"
                + ",".join(f"{host.address}:{host.port}" for host in config.connection.hosts)
                + f"/?replicaSet={config.connection.replica_set}"
            )
            return pymongo.MongoClient(
                uri,
                replicaSet=config.connection.replica_set,
                connectTimeoutMS=3000,
                serverSelectionTimeoutMS=5000,
            )
        else:
            raise NotImplementedError()

    @staticmethod
    def _master_col_name(entity: str) -> str:
        """Returns name of master collection for `entity`."""
        return f"{entity}#master"

    @staticmethod
    def _snapshots_col_name(entity: str) -> str:
        """Returns name of snapshots collection for `entity`."""
        return f"{entity}#snapshots"

    @staticmethod
    def _raw_col_name(entity: str) -> str:
        """Returns name of raw data collection for `entity`."""
        return f"{entity}#raw"

    def _init_database_schema(self, db_name) -> None:
        """Runs full check and update of database schema.

        Checks whether `db_name` database exists. If not, creates it.
        Checks whether all collections exist. If not, creates them.
        This all is done automatically in MongoDB.

        We need to create indexes, however.
        """
        for etype in self._db_schema_config.entities:
            # Snapshots index on `eid` and `_time_created` fields
            snapshot_col = self._snapshots_col_name(etype)
            self._db[snapshot_col].create_index("eid", background=True)
            self._db[snapshot_col].create_index("_time_created", background=True)

            # Existence indexes for each attribute in master collection
            master_col = self._master_col_name(etype)
            for attr, _spec in self._db_schema_config.entity_attributes[etype].items():
                self._db[master_col].create_index(
                    attr, background=True, partialFilterExpression={attr: {"$exists": True}}
                )

    def _assert_etype_exists(self, etype: str):
        """Asserts `etype` existence.

        If doesn't exist, raises `DatabaseError`.
        """
        if etype not in self._db_schema_config.entities:
            raise DatabaseError(f"Entity '{etype}' does not exist")

    def insert_datapoints(
        self, etype: str, eid: str, dps: list[DataPointBase], new_entity: bool = False
    ) -> None:
        """Inserts datapoint to raw data collection and updates master record.

        Raises DatabaseError when insert or update fails.
        """
        if len(dps) == 0:
            return

        etype = dps[0].etype

        # Check `etype`
        self._assert_etype_exists(etype)

        # Insert raw datapoints
        raw_col = self._raw_col_name(etype)
        dps_dicts = [dp.dict(exclude={"attr_type"}) for dp in dps]
        try:
            self._db[raw_col].insert_many(dps_dicts)
            self.log.debug(f"Inserted datapoints to raw collection:\n{dps}")
        except Exception as e:
            raise DatabaseError(f"Insert of datapoints failed: {e}\n{dps}") from e

        # Update master document
        master_changes = {"$push": {}, "$set": {}}
        for dp in dps:
            attr_spec = self._db_schema_config.attr(etype, dp.attr)

            v = dp.v.dict() if isinstance(dp.v, BaseModel) else dp.v

            # Rewrite value of plain attribute
            if attr_spec.t == AttrType.PLAIN:
                master_changes["$set"][dp.attr] = {"v": v, "ts_last_update": datetime.now()}

            # Push new data of observation
            if attr_spec.t == AttrType.OBSERVATIONS:
                if dp.attr in master_changes["$push"]:
                    # Support multiple datapoints being pushed in the same request
                    if "$each" not in master_changes["$push"][dp.attr]:
                        saved_dp = master_changes["$push"][dp.attr]
                        master_changes["$push"][dp.attr] = {"$each": [saved_dp]}
                    master_changes["$push"][dp.attr]["$each"].append(
                        {"t1": dp.t1, "t2": dp.t2, "v": v, "c": dp.c}
                    )
                else:
                    # Otherwise just push one datapoint
                    master_changes["$push"][dp.attr] = {"t1": dp.t1, "t2": dp.t2, "v": v, "c": dp.c}

            # Push new data of timeseries
            if attr_spec.t == AttrType.TIMESERIES:
                if dp.attr in master_changes["$push"]:
                    # Support multiple datapoints being pushed in the same request
                    if "$each" not in master_changes["$push"][dp.attr]:
                        saved_dp = master_changes["$push"][dp.attr]
                        master_changes["$push"][dp.attr] = {"$each": [saved_dp]}
                    master_changes["$push"][dp.attr]["$each"].append(
                        {"t1": dp.t1, "t2": dp.t2, "v": v}
                    )
                else:
                    # Otherwise just push one datapoint
                    master_changes["$push"][dp.attr] = {"t1": dp.t1, "t2": dp.t2, "v": v}

        if new_entity:
            master_changes["$set"]["#hash"] = HASH(f"{etype}:{eid}")

        master_col = self._master_col_name(etype)
        try:
            self._db[master_col].update_one({"_id": eid}, master_changes, upsert=True)
            self.log.debug(f"Updated master record of {etype} {eid}: {master_changes}")
        except Exception as e:
            raise DatabaseError(f"Update of master record failed: {e}\n{dps}") from e

    def update_master_records(self, etype: str, eids: list[str], records: list[dict]) -> None:
        """Replace master records of `etype`:`eid` with the provided `records`.

        Raises DatabaseError when update fails.
        """
        master_col = self._master_col_name(etype)
        try:
            self._db[master_col].bulk_write(
                [
                    ReplaceOne({"_id": eid}, record, upsert=True)
                    for eid, record in zip(eids, records)
                ]
            )
            self.log.debug("Updated master records of %s: %s.", eids, eids)
        except Exception as e:
            raise DatabaseError(f"Update of master records failed: {e}\n{records}") from e

    def delete_old_dps(self, etype: str, attr_name: str, t_old: datetime) -> None:
        """Delete old datapoints from master collection.

        Periodically called for all `etype`s from HistoryManager.
        """
        master_col = self._master_col_name(etype)
        try:
            self._db[master_col].update_many({}, {"$pull": {attr_name: {"t2": {"$lt": t_old}}}})
        except Exception as e:
            raise DatabaseError(f"Delete of old datapoints failed: {e}") from e

    def get_master_record(self, etype: str, eid: str, **kwargs) -> dict:
        """Get current master record for etype/eid.

        If doesn't exist, returns {}.
        """
        # Check `etype`
        self._assert_etype_exists(etype)

        master_col = self._master_col_name(etype)
        return self._db[master_col].find_one({"_id": eid}, **kwargs) or {}

    def ekey_exists(self, etype: str, eid: str) -> bool:
        """Checks whether master record for etype/eid exists"""
        return bool(self.get_master_record(etype, eid))

    def get_master_records(self, etype: str, **kwargs) -> pymongo.cursor.Cursor:
        """Get cursor to current master records of etype."""
        # Check `etype`
        self._assert_etype_exists(etype)

        master_col = self._master_col_name(etype)
        return self._db[master_col].find({}, **kwargs)

    def get_worker_master_records(
        self, worker_index: int, worker_cnt: int, etype: str, **kwargs
    ) -> pymongo.cursor.Cursor:
        """Get cursor to current master records of etype."""
        if etype not in self._db_schema_config.entities:
            raise DatabaseError(f"Entity '{etype}' does not exist")

        master_col = self._master_col_name(etype)
        return self._db[master_col].find({"#hash": {"$mod": [worker_cnt, worker_index]}}, **kwargs)

    def get_latest_snapshot(self, etype: str, eid: str) -> dict:
        """Get latest snapshot of given etype/eid.

        If doesn't exist, returns {}.
        """
        # Check `etype`
        self._assert_etype_exists(etype)

        snapshot_col = self._snapshots_col_name(etype)
        return self._db[snapshot_col].find_one({"eid": eid}, sort=[("_id", -1)]) or {}

    def get_latest_snapshots(
        self, etype: str, eid_filter: str = ""
    ) -> tuple[pymongo.cursor.Cursor, int]:
        """Get latest snapshots of given `etype`.

        This method is useful for displaying data on web.

        If `eid_filter` is not empty, returns only `eid`s containing substring `eid_filter`.

        Also returns total document count (after filtering).

        May raise `DatabaseError` if query is invalid.
        """
        # Check `etype`
        self._assert_etype_exists(etype)

        snapshot_col = self._snapshots_col_name(etype)
        latest_snapshot = self._db[snapshot_col].find_one({}, sort=[("_id", -1)])
        if latest_snapshot is None:
            return self._db[snapshot_col].find(), self._db[snapshot_col].count_documents({})

        latest_snapshot_date = latest_snapshot["_time_created"]
        query = {"_time_created": latest_snapshot_date}
        if eid_filter != "":
            query["eid"] = {"$regex": eid_filter}

        try:
            return self._db[snapshot_col].find(query), self._db[snapshot_col].count_documents(query)
        except OperationFailure as e:
            raise DatabaseError("Invalid query") from e

    def get_snapshots(
        self, etype: str, eid: str, t1: Optional[datetime] = None, t2: Optional[datetime] = None
    ) -> pymongo.cursor.Cursor:
        """Get all (or filtered) snapshots of given `eid`.

        This method is useful for displaying `eid`'s history on web.

        Args:
            etype: entity type
            eid: id of entity, to which data-points correspond
            t1: left value of time interval (inclusive)
            t2: right value of time interval (inclusive)
        """
        # Check `etype`
        self._assert_etype_exists(etype)

        snapshot_col = self._snapshots_col_name(etype)
        query = {"eid": eid, "_time_created": {}}

        # Filter by date
        if t1:
            query["_time_created"]["$gte"] = t1
        if t2:
            query["_time_created"]["$lte"] = t2

        # Unset if empty
        if not query["_time_created"]:
            del query["_time_created"]

        return self._db[snapshot_col].find(query).sort([("_time_created", pymongo.ASCENDING)])

    def get_value_or_history(
        self,
        etype: str,
        attr_name: str,
        eid: str,
        t1: Optional[datetime] = None,
        t2: Optional[datetime] = None,
    ) -> dict:
        """Gets current value and/or history of attribute for given `eid`.

        Depends on attribute type:
        - plain: just (current) value
        - observations: (current) value and history stored in master record (optionally filtered)
        - timeseries: just history stored in master record (optionally filtered)

        Returns dict with two keys: `current_value` and `history` (list of values).
        """
        # Check `etype`
        self._assert_etype_exists(etype)

        attr_spec = self._db_schema_config.attr(etype, attr_name)

        result = {"current_value": None, "history": []}

        # Add current value to the result
        if attr_spec.t == AttrType.PLAIN:
            result["current_value"] = (
                self.get_master_record(etype, eid).get(attr_name, {}).get("v", None)
            )
        elif attr_spec.t == AttrType.OBSERVATIONS:
            result["current_value"] = self.get_latest_snapshot(etype, eid).get(attr_name, None)

        # Add history
        if attr_spec.t == AttrType.OBSERVATIONS:
            result["history"] = self.get_observation_history(etype, attr_name, eid, t1, t2)
        elif attr_spec.t == AttrType.TIMESERIES:
            result["history"] = self.get_timeseries_history(etype, attr_name, eid, t1, t2)

        return result

    def estimate_count_eids(self, etype: str) -> int:
        """Estimates count of `eid`s in given `etype`"""
        # Check `etype`
        self._assert_etype_exists(etype)

        master_col = self._master_col_name(etype)
        return self._db[master_col].estimated_document_count({})

    def save_snapshot(self, etype: str, snapshot: dict, time: datetime):
        """Saves snapshot to specified entity of current master document."""
        # Check `etype`
        self._assert_etype_exists(etype)

        snapshot["_time_created"] = time

        snapshot_col = self._snapshots_col_name(etype)
        try:
            self._db[snapshot_col].insert_one(snapshot)
            self.log.debug(f"Inserted snapshot: {snapshot}")
        except Exception as e:
            raise DatabaseError(f"Insert of snapshot failed: {e}\n{snapshot}") from e

    def save_snapshots(self, etype: str, snapshots: list[dict], time: datetime):
        """
        Saves a list of snapshots of current master documents.

        All snapshots must belong to same entity type.
        """
        # Check `etype`
        self._assert_etype_exists(etype)

        for snapshot in snapshots:
            snapshot["_time_created"] = time

        snapshot_col = self._snapshots_col_name(etype)
        try:
            self._db[snapshot_col].insert_many(snapshots)
            self.log.debug(f"Inserted snapshots: {snapshots}")
        except Exception as e:
            raise DatabaseError(f"Insert of snapshots failed: {e}\n{snapshots}") from e

    def save_metadata(self, time: datetime, metadata: dict):
        """Saves snapshot to specified entity of current master document."""
        module = get_caller_id()
        metadata["_id"] = module + time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]
        metadata["#module"] = module
        metadata["#time_created"] = time
        metadata["#last_update"] = datetime.now()
        try:
            self._db["#metadata"].insert_one(metadata)
            self.log.debug("Inserted metadata %s: %s", metadata["_id"], metadata)
        except Exception as e:
            raise DatabaseError(f"Insert of metadata failed: {e}\n{metadata}") from e

    def update_metadata(self, time: datetime, metadata: dict, increase: dict = None):
        module = get_caller_id()
        metadata_id = module + time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]
        metadata["#last_update"] = datetime.now()

        changes = {"$set": metadata} if increase is None else {"$set": metadata, "$inc": increase}

        try:
            res = self._db["#metadata"].update_one({"_id": metadata_id}, changes)
            self.log.debug(
                "Updated metadata %s, changes: %s, result: %s", metadata_id, changes, res.raw_result
            )
        except Exception as e:
            raise DatabaseError(f"Update of metadata failed: {e}\n{metadata_id}, {changes}") from e

    def get_observation_history(
        self,
        etype: str,
        attr_name: str,
        eid: str,
        t1: datetime = None,
        t2: datetime = None,
        sort: int = None,
    ) -> list[dict]:
        """Get full (or filtered) history of observation attribute.

        This method is useful for displaying `eid`'s history on web.
        Also used to feed data into `get_timeseries_history()`.

        Args:
            etype: entity type
            attr_name: name of attribute
            eid: id of entity, to which data-points correspond
            t1: left value of time interval (inclusive)
            t2: right value of time interval (inclusive)
            sort: sort by timestamps - 0: ascending order by t1, 1: descending order by t2,
                None: don't sort
        Returns:
            list of dicts (reduced datapoints)
        """
        t1 = datetime.fromtimestamp(0) if t1 is None else t1.replace(tzinfo=None)
        t2 = datetime.now() if t2 is None else t2.replace(tzinfo=None)

        # Get attribute history
        mr = self.get_master_record(etype, eid)
        attr_history = mr.get(attr_name, [])

        # Filter
        attr_history_filtered = [row for row in attr_history if row["t1"] <= t2 and row["t2"] >= t1]

        # Sort
        if sort == 1:
            attr_history_filtered.sort(key=lambda row: row["t1"])
        elif sort == 2:
            attr_history_filtered.sort(key=lambda row: row["t2"], reverse=True)

        return attr_history_filtered

    def get_timeseries_history(
        self,
        etype: str,
        attr_name: str,
        eid: str,
        t1: datetime = None,
        t2: datetime = None,
        sort: int = None,
    ) -> list[dict]:
        """Get full (or filtered) history of timeseries attribute.
        Outputs them in format:
        ```
            [
                {
                    "t1": ...,
                    "t2": ...,
                    "v": {
                        "series1": ...,
                        "series2": ...
                    }
                },
                ...
            ]
        ```
        This method is useful for displaying `eid`'s history on web.

        Args:
            etype: entity type
            attr_name: name of attribute
            eid: id of entity, to which data-points correspond
            t1: left value of time interval (inclusive)
            t2: right value of time interval (inclusive)
            sort: sort by timestamps - `0`: ascending order by `t1`, `1`: descending order by `t2`,
                `None`: don't sort
        Returns:
             list of dicts (reduced datapoints) - each represents just one point at time
        """
        t1 = datetime.fromtimestamp(0) if t1 is None else t1.replace(tzinfo=None)
        t2 = datetime.now() if t2 is None else t2.replace(tzinfo=None)

        attr_history = self.get_observation_history(etype, attr_name, eid, t1, t2, sort)
        if not attr_history:
            return []

        attr_history_split = self._split_timeseries_dps(etype, attr_name, attr_history)

        # Filter out rows outside [t1, t2] interval
        attr_history_filtered = [
            row for row in attr_history_split if row["t1"] <= t2 and row["t2"] >= t1
        ]

        return attr_history_filtered

    def _split_timeseries_dps(
        self, etype: str, attr_name: str, attr_history: list[dict]
    ) -> list[dict]:
        """Helper to split "datapoints" (rows) of timeseries to "datapoints"
        containing just one value per series."""
        attrib_conf = self._db_schema_config.attr(etype, attr_name)
        timeseries_type = attrib_conf.timeseries_type

        result = []

        # Should be ["time"] or ["time_first", "time_last"] for irregular
        # timeseries and [] for regular timeseries
        time_series = list(timeseries_types[timeseries_type]["default_series"].keys())

        # User-defined series
        user_series = list(set(attrib_conf.series.keys()) - set(time_series))

        if not user_series:
            return []

        if timeseries_type == "regular":
            time_step = attrib_conf.timeseries_params.time_step

            for row in attr_history:
                # Length of all series arrays/lists in this row
                values_len = len(row["v"][user_series[0]])

                for i in range(values_len):
                    row_t1 = row["t1"] + i * time_step
                    row_t2 = row_t1 + time_step

                    row_new = {"t1": row_t1, "t2": row_t2, "v": {}}

                    for s in user_series:
                        row_new["v"][s] = row["v"][s][i]

                    result.append(row_new)
        else:
            for row in attr_history:
                # Length of all series arrays/lists in this row
                values_len = len(row["v"][user_series[0]])

                for i in range(values_len):
                    row_t1 = row["v"].get(time_series[0])[i] if len(time_series) >= 1 else None
                    row_t2 = row["v"].get(time_series[1])[i] if len(time_series) >= 2 else row_t1

                    row_new = {"t1": row_t1, "t2": row_t2, "v": {}}

                    for s in user_series:
                        row_new["v"][s] = row["v"][s][i]

                    result.append(row_new)

        return result

    def get_raw_summary(self, etype: str, before: datetime) -> pymongo.cursor:
        raw_col_name = self._raw_col_name(etype)
        try:
            return self._db[raw_col_name].aggregate(
                [
                    {"$match": {"t1": {"$lt": before}}},
                    {
                        "$group": {
                            "_id": "date_range",
                            "earliest": {"$min": "$t1"},
                            "latest": {"$max": "$t1"},
                            "count": {"$sum": 1},
                        }
                    },
                ]
            )
        except Exception as e:
            raise DatabaseError(f"Raw collection {raw_col_name} fetch failed: {e}") from e

    def get_raw(
        self, etype: str, after: datetime, before: datetime, plain: bool = True
    ) -> pymongo.cursor:
        """Get raw datapoints where `t1` is in <`after`, `before`).

        If `plain` is `True`, then all plain datapoints will be returned (default).
        """
        raw_col_name = self._raw_col_name(etype)
        query_filter = {"$or": [{"t1": {"$gte": after, "$lt": before}}]}
        if plain:
            query_filter["$or"].append({"t1": {"$exists": False}})
        return self._db[raw_col_name].find(query_filter)

    def delete_old_raw_dps(self, etype: str, before: datetime, plain: bool = True):
        """Delete raw datapoints older than `before`.

        Deletes all plain datapoints if `plain` is `True` (default).
        """
        raw_col_name = self._raw_col_name(etype)
        query_filter = {"$or": [{"t1": {"$lt": before}}]}
        if plain:
            query_filter["$or"].append({"t1": {"$exists": False}})
        try:
            return self._db[raw_col_name].delete_many(query_filter)
        except Exception as e:
            raise DatabaseError(f"Delete of old datapoints failed: {e}") from e

    def delete_old_snapshots(self, etype: str, t_old: datetime):
        """Delete old snapshots.

        Periodically called for all `etype`s from HistoryManager.
        """
        snapshot_col_name = self._snapshots_col_name(etype)
        try:
            return self._db[snapshot_col_name].delete_many({"_time_created": {"$lt": t_old}})
        except Exception as e:
            raise DatabaseError(f"Delete of olds snapshots failed: {e}") from e

    def get_module_cache(self):
        """Return a persistent cache collection for given module name."""
        module = get_caller_id()
        self.log.debug("Module %s is accessing its cache collection", module)
        return self._db[f"#cache#{module}"]
